use std::{
    collections::{hash_map::Entry, HashMap},
    hash::Hash,
    sync::Mutex,
};

use egui::{
    epaint::ImageDelta,
    load::{LoadError, SizedTexture, TextureLoader, TexturePoll},
    ColorImage, Context, TextureHandle, TextureOptions,
};
use serde::{Deserialize, Serialize};

pub trait VramResource: Sized + PartialEq + Eq + Hash {
    fn to_uri(&self) -> String;
    fn from_uri(uri: &str) -> Option<Self>;
}

impl<T: Serialize + for<'a> Deserialize<'a> + PartialEq + Eq + Hash> VramResource for T {
    fn to_uri(&self) -> String {
        format!("vram://{}", serde_json::to_string(self).unwrap())
    }

    fn from_uri(uri: &str) -> Option<Self> {
        let content = uri.strip_prefix("vram://")?;
        serde_json::from_str(content).ok()
    }
}

pub trait VramImageLoader {
    type Resource: VramResource;

    fn id(&self) -> &str;
    fn add(&self, resource: &Self::Resource) -> Option<ColorImage>;
    fn update<'a>(
        &'a self,
        resources: impl Iterator<Item = &'a Self::Resource>,
    ) -> Vec<(&'a Self::Resource, ColorImage)>;
}

pub struct VramTextureLoader<T: VramImageLoader> {
    id: String,
    loader: Mutex<T>,
    cache: Mutex<HashMap<T::Resource, TextureHandle>>,
}

impl<T: VramImageLoader> VramTextureLoader<T> {
    pub fn new(loader: T) -> Self {
        Self {
            id: loader.id().to_string(),
            loader: Mutex::new(loader),
            cache: Mutex::new(HashMap::new()),
        }
    }
}

impl<T: VramImageLoader> TextureLoader for VramTextureLoader<T> {
    fn id(&self) -> &str {
        &self.id
    }

    fn load(
        &self,
        ctx: &Context,
        uri: &str,
        texture_options: TextureOptions,
        _size_hint: egui::SizeHint,
    ) -> Result<TexturePoll, LoadError> {
        let Some(resource) = T::Resource::from_uri(uri) else {
            return Err(LoadError::NotSupported);
        };
        if texture_options != TextureOptions::NEAREST {
            return Err(LoadError::Loading(
                "Only TextureOptions::NEAREST are supported".into(),
            ));
        }
        let loader = self.loader.lock().unwrap();
        let mut cache = self.cache.lock().unwrap();
        for (resource, updated_image) in loader.update(cache.keys()) {
            if let Some(handle) = cache.get(resource) {
                let delta = ImageDelta::full(updated_image, TextureOptions::NEAREST);
                ctx.tex_manager().write().set(handle.id(), delta);
            }
        }
        match cache.entry(resource) {
            Entry::Occupied(entry) => {
                let texture = SizedTexture::from_handle(entry.get());
                Ok(TexturePoll::Ready { texture })
            }
            Entry::Vacant(entry) => {
                if let Some(image) = loader.add(entry.key()) {
                    let handle =
                        entry.insert(ctx.load_texture(uri, image, TextureOptions::NEAREST));
                    let texture = SizedTexture::from_handle(handle);
                    Ok(TexturePoll::Ready { texture })
                } else {
                    Err(LoadError::Loading("could not load texture".into()))
                }
            }
        }
    }

    fn forget(&self, uri: &str) {
        if let Some(resource) = T::Resource::from_uri(uri) {
            self.cache.lock().unwrap().remove(&resource);
        }
    }

    fn forget_all(&self) {
        self.cache.lock().unwrap().clear();
    }

    fn byte_size(&self) -> usize {
        self.cache
            .lock()
            .unwrap()
            .values()
            .map(|h| h.byte_size())
            .sum()
    }
}
