use std::sync::Arc;

use egui::{CentralPanel, ColorImage, Context, Image, TextureOptions, ViewportBuilder, ViewportId};
use serde::{Deserialize, Serialize};

use crate::{
    emulator::SimId,
    memory::{MemoryMonitor, MemoryView},
    vram::{VramImageLoader, VramResource as _, VramTextureLoader},
    window::AppWindow,
};

use super::utils::parse_palette;

pub struct BgMapWindow {
    sim_id: SimId,
    loader: Option<BgMapLoader>,
}

impl BgMapWindow {
    pub fn new(sim_id: SimId, memory: &mut MemoryMonitor) -> Self {
        Self {
            sim_id,
            loader: Some(BgMapLoader::new(sim_id, memory)),
        }
    }
}

impl AppWindow for BgMapWindow {
    fn viewport_id(&self) -> ViewportId {
        ViewportId::from_hash_of(format!("bgmap-{}", self.sim_id))
    }

    fn sim_id(&self) -> SimId {
        self.sim_id
    }

    fn initial_viewport(&self) -> ViewportBuilder {
        ViewportBuilder::default()
            .with_title(format!("BG Map Data ({})", self.sim_id))
            .with_inner_size((640.0, 480.0))
    }

    fn on_init(&mut self, ctx: &Context, _render_state: &egui_wgpu::RenderState) {
        let loader = self.loader.take().unwrap();
        ctx.add_texture_loader(Arc::new(VramTextureLoader::new(loader)));
    }

    fn show(&mut self, ctx: &Context) {
        CentralPanel::default().show(ctx, |ui| {
            let resource = BgMapResource { index: 0 };
            let image = Image::new(resource.to_uri()).texture_options(TextureOptions::NEAREST);
            ui.add(image);
        });
    }
}

#[derive(Serialize, Deserialize, PartialEq, Eq, Hash)]
struct BgMapResource {
    index: usize,
}

struct BgMapLoader {
    chardata: MemoryView,
    bgmaps: MemoryView,
    brightness: MemoryView,
    palettes: MemoryView,
}

impl BgMapLoader {
    pub fn new(sim_id: SimId, memory: &mut MemoryMonitor) -> Self {
        Self {
            chardata: memory.view(sim_id, 0x00078000, 0x8000),
            bgmaps: memory.view(sim_id, 0x00020000, 0x1d800),
            brightness: memory.view(sim_id, 0x0005f824, 8),
            palettes: memory.view(sim_id, 0x0005f860, 16),
        }
    }

    fn load_bgmap(&self, index: usize) -> Option<ColorImage> {
        let chardata = self.chardata.borrow();
        let bgmaps = self.bgmaps.borrow();
        let brightness = self.brightness.borrow();
        let palettes = self.palettes.borrow();

        let brts = brightness.range::<u8>(0, 8);
        let colors = [
            parse_palette(palettes.read(0), brts),
            parse_palette(palettes.read(2), brts),
            parse_palette(palettes.read(4), brts),
            parse_palette(palettes.read(6), brts),
        ];

        let mut data = vec![0u8; 512 * 512];
        for (i, cell) in bgmaps.range::<u16>(index * 4096, 4096).iter().enumerate() {
            let char_index = (cell & 0x7ff) as usize;
            let char = chardata.range::<u16>(char_index * 8, 8);
            let vflip = cell & 0x1000 != 0;
            let hflip = cell & 0x2000 != 0;
            let palette_index = (cell >> 14) as usize;
            let palette = &colors[palette_index];

            let mut target_idx = (i % 64) * 8 + (i / 64) * 8 * 512;
            for row in 0..8 {
                let dests = &mut data[target_idx..target_idx + 8];
                let pixels = self.read_char_row(char, hflip, vflip, row);
                for (dest, pixel) in dests.iter_mut().zip(pixels) {
                    *dest = palette[pixel as usize];
                }
                target_idx += 512;
            }
        }

        Some(ColorImage::from_gray([512, 512], &data))
    }

    fn read_char_row(
        &self,
        char: &[u16],
        hflip: bool,
        vflip: bool,
        row: usize,
    ) -> impl Iterator<Item = u8> {
        let pixels = if vflip { char[7 - row] } else { char[row] };
        (0..16).step_by(2).map(move |i| {
            let pixel = if hflip { 14 - i } else { i };
            ((pixels >> pixel) & 0x3) as u8
        })
    }
}

impl VramImageLoader for BgMapLoader {
    type Resource = BgMapResource;

    fn id(&self) -> &str {
        concat!(module_path!(), "::BgMapLoader")
    }

    fn add(&self, resource: &Self::Resource) -> Option<ColorImage> {
        let BgMapResource { index } = resource;
        self.load_bgmap(*index)
    }

    fn update<'a>(
        &'a self,
        resources: impl Iterator<Item = &'a Self::Resource>,
    ) -> Vec<(&'a Self::Resource, ColorImage)> {
        let _ = resources;
        vec![]
    }
}
